/*
  Windows 2000 XP API Wrapper Pack
  Copyright (C) 2008 OldCigarette

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <windows.h>
#include <shlobj.h>
#include <shobjidl.h>
#include <shlwapi.h>
#include "../common/win2k_xp_config.h"

/*Debugging*/
#include "../common/win2k_xp_debug.h"
extern "C" extern int debugLevel;
extern "C" extern DbgPrintf_t DbgPrintf;

/* The IShellItem COM object is completely missing and needs implemented.  I
   think most of the functionality can come from shellfolder
HRESULT WINAPI XP_SHCreateShellItem(PVOID pidlParent, IShellFolder *psfParent,
  PVOID pidl, IShellItem **ppsi) {
	return SHCreateShellItem(pidlParent, psfParent, pidl, ppsi);
}*/

struct ShellItem : public IShellItem {
  public:
	virtual HRESULT STDMETHODCALLTYPE BindToHandler( 
            /* [in] */ IBindCtx *pbc,
            /* [in] */ REFGUID rbhid,
            /* [in] */ REFIID riid,
            /* [iid_is][out] */ void **ppvOut);

    virtual HRESULT STDMETHODCALLTYPE GetParent( 
            /* [out] */ IShellItem **ppsi);
        
    virtual HRESULT STDMETHODCALLTYPE GetDisplayName( 
            /* [in] */ SIGDN sigdnName,
            /* [string][out] */ LPWSTR *ppszName);
        
    virtual HRESULT STDMETHODCALLTYPE GetAttributes( 
            /* [in] */ SFGAOF sfgaoMask,
            /* [out] */ SFGAOF *psfgaoAttribs);
        
    virtual HRESULT STDMETHODCALLTYPE Compare( 
            /* [in] */ IShellItem *psi,
            /* [in] */ SICHINTF hint,
            /* [out] */ int *piOrder);

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(
			REFIID iid,
			void ** ppvObject);
	
	virtual ULONG STDMETHODCALLTYPE AddRef(void);
	virtual ULONG STDMETHODCALLTYPE Release(void);
	
	ShellItem(IShellFolder *parentFolder, LPITEMIDLIST itemPidl);
	~ShellItem();
	LPITEMIDLIST GetAbsoluteIDL();
	
	IShellFolder *parentFolder;
	LPITEMIDLIST  itemPidl;
	ULONG ref_count;
};

ULONG STDMETHODCALLTYPE ShellItem::AddRef(void) {
	ref_count++;
	return ref_count;
}

ULONG STDMETHODCALLTYPE ShellItem::Release(void) {
	ULONG pass_back;
	ref_count--;
	if(ref_count <= 0) {
		pass_back = ref_count;
		delete this;
		return pass_back;
	}
	return ref_count;
}

HRESULT STDMETHODCALLTYPE ShellItem::QueryInterface(REFIID iid, void ** ppvObject) {
	return E_NOINTERFACE;
}

ShellItem::ShellItem(IShellFolder *p, LPITEMIDLIST i) :
  parentFolder(p), itemPidl(i) {
}

ShellItem::~ShellItem() {
	parentFolder->Release();
	ILFree(itemPidl);
}

LPITEMIDLIST ShellItem::GetAbsoluteIDL() {
	LPITEMIDLIST absItemPidl;
	SHGetRealIDL(parentFolder, itemPidl, &absItemPidl);
	if(absItemPidl) absItemPidl = ILClone(absItemPidl);
	return absItemPidl;
}

DWORD SIGDNtoSHGNO(SIGDN flags) {
	//By all means lets create another enum for ShellItem to do the same fucking thing
	if(flags == SIGDN_PARENTRELATIVEEDITING)       return SHGDN_INFOLDER|SHGDN_FOREDITING;
	if(flags == SIGDN_PARENTRELATIVEFORADDRESSBAR) return SHGDN_INFOLDER|SHGDN_FORADDRESSBAR;
	if(flags == SIGDN_PARENTRELATIVEPARSING)       return SHGDN_INFOLDER|SHGDN_FORPARSING;
	if(flags == SIGDN_DESKTOPABSOLUTEEDITING)      return SHGDN_FOREDITING;
	if(flags == SIGDN_DESKTOPABSOLUTEPARSING)      return SHGDN_FORPARSING;
	if(flags == SIGDN_NORMALDISPLAY)               return SHGDN_NORMAL;
	if(flags == SIGDN_FILESYSPATH)                 return SHGDN_FORPARSING;
	if(flags == SIGDN_URL)                         return SHGDN_FORADDRESSBAR;
	return SHGDN_NORMAL;
}

HRESULT STDMETHODCALLTYPE ShellItem::GetDisplayName(SIGDN sigdnName, LPWSTR *ppszName) {
	HRESULT hr;
	STRRET sr;
	DWORD flags = SIGDNtoSHGNO(sigdnName);
	
	sr.uType = STRRET_WSTR;
	if((hr = parentFolder->GetDisplayNameOf(itemPidl, flags, &sr)) == S_OK) {
		StrRetToStrW(&sr, itemPidl, ppszName);
		CoTaskMemFree(sr.pOleStr);
		return S_OK;
	} else {
		return hr;
	}
}

HRESULT STDMETHODCALLTYPE ShellItem::GetParent(IShellItem **ppsi) {
	LPITEMIDLIST absItemPidl;
	LPCITEMIDLIST itemPidl;
	LPITEMIDLIST newitemPidl;
	IShellFolder  *newparentFolder = NULL;
	LPMALLOC ppMalloc;
	
	absItemPidl = GetAbsoluteIDL();
	if(absItemPidl) {
		ILRemoveLastID(absItemPidl);
		if(!SUCCEEDED(SHBindToParent(absItemPidl, IID_IShellFolder, (void **)&newparentFolder, &itemPidl))) {
			ILFree(absItemPidl);
			return E_FAIL;
		}
		newitemPidl = ILClone(itemPidl);
		ILFree(absItemPidl);
		*ppsi = new ShellItem(newparentFolder, newitemPidl);
		(*ppsi)->AddRef();
		return S_OK;
	} else {
		return E_FAIL;
	}
}

HRESULT STDMETHODCALLTYPE ShellItem::GetAttributes(SFGAOF sfgaoMask, SFGAOF *psfgaoAttribs) {
	HRESULT hr;
	LPCITEMIDLIST idarray[1];
	idarray[0] = itemPidl;
	if((hr = parentFolder->GetAttributesOf(1, idarray, &sfgaoMask)) == S_OK) {
		*psfgaoAttribs = sfgaoMask;
		return S_OK;
	} else {
		return hr;
	}
}

HRESULT STDMETHODCALLTYPE ShellItem::Compare(IShellItem *psi, SICHINTF hint, int *piOrder) {
	HRESULT hr;
	IShellFolder *desktopFolder;
	LPITEMIDLIST absThis;
	LPITEMIDLIST absOther;
	
	if(SHGetDesktopFolder(&desktopFolder) != S_OK) return E_FAIL;
	
	absThis  = GetAbsoluteIDL();
	if(absThis == NULL) { desktopFolder->Release(); return E_FAIL; }
	
	absOther = ((ShellItem *)psi)->GetAbsoluteIDL();
	if(absOther == NULL) { desktopFolder->Release(); ILFree(absThis); return E_FAIL; }
	
	hr = desktopFolder->CompareIDs(hint, absThis, absOther);
	desktopFolder->Release();
	
	*piOrder = HRESULT_CODE(hr);
	if(*piOrder == 0) return S_OK;    //same
	else              return S_FALSE; //different
}

HRESULT STDMETHODCALLTYPE ShellItem::BindToHandler(IBindCtx *pbc, REFGUID rbhid, REFIID riid, void **ppvOut) {
	if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "shell32: BindToHandler not implemented\n");
	return E_FAIL;
}

extern "C" {
HRESULT WINAPI XP_SHCreateShellItem(LPCITEMIDLIST pidlParent, IShellFolder *psfParent,
  LPCITEMIDLIST pidl, IShellItem **ppsi) {
	LPITEMIDLIST absItemPidl = NULL;
	LPITEMIDLIST itemPidl = NULL;
	IShellFolder  *parentFolder = NULL;
	LPMALLOC ppMalloc;
	
	if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "shell32: SHCreateShellItem partial implementation\n");
	
	if(pidlParent == NULL && psfParent == NULL) {
		//The item is identified by an absolute IDList passed to the pidl parameter.
		absItemPidl = ILClone(pidl);
	} else if(pidlParent == NULL) {
		//The parent folder is identified by psfParent. The pidl parameter points to a child IDList that identifies the item within the folder identified by psfParent.
		parentFolder = psfParent;
		parentFolder->AddRef();
		itemPidl = ILClone(pidl);
	} else if(psfParent == NULL) {
		//The parent folder is identified by an absolute IDList pidlParent. The pidl parameter points to a child IDList that identifies the item within the folder identified by pidlParent.
		absItemPidl = ILCombine(pidlParent, pidl);
	}
	
	if(absItemPidl) {
		LPCITEMIDLIST citemPidl = NULL;
		if(!SUCCEEDED(SHBindToParent(absItemPidl, IID_IShellFolder, (void **)&parentFolder, &citemPidl))) {
			ILFree(absItemPidl);
			return E_FAIL;
		}
		itemPidl = ILClone(citemPidl);
		ILFree(absItemPidl);
	}
		
	if(itemPidl == NULL || parentFolder == NULL) return E_FAIL;

	//const_cast is all SHBindToParent's fault
	*ppsi = new ShellItem(parentFolder, itemPidl);
	return S_OK;
}
};