/*
  Windows 2000 XP API Wrapper Pack
  Copyright (C) 2008 OldCigarette

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#define DIRECTINPUT_VERSION 0x0800
#define DEFINE_RAW_INPUT
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <dinput.h>
#include <stdio.h>

#define WM_INPUT            0x00FF

#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED           0x00080000
#define LWA_COLORKEY            0x00000001
#define LWA_ALPHA               0x00000002
#endif // ndef WS_EX_LAYERED

#define RECTWIDTH(r) ((r).right - (r).left)
#define RECTHEIGHT(r) ((r).bottom - (r).top)

#ifndef PW_CLIENTONLY
#define PW_CLIENTONLY 0x00000001
#endif


typedef struct _MOUSE_ATTRIBUTES {
    USHORT MouseIdentifier;
    USHORT NumberOfButtons;
    USHORT SampleRate;
    ULONG  InputDataQueueLength;
} MOUSE_ATTRIBUTES, *PMOUSE_ATTRIBUTES;

typedef struct _KEYBOARD_TYPEMATIC_PARAMETERS {
    USHORT  UnitId;
    USHORT  Rate;
    USHORT  Delay;
} KEYBOARD_TYPEMATIC_PARAMETERS, *PKEYBOARD_TYPEMATIC_PARAMETERS;

typedef struct _KEYBOARD_ID {
    UCHAR Type;       // Keyboard type
    UCHAR Subtype;    // Keyboard subtype (OEM-dependent value)
} KEYBOARD_ID, *PKEYBOARD_ID;

typedef struct _KEYBOARD_ATTRIBUTES {
  KEYBOARD_ID  KeyboardIdentifier;
  USHORT  KeyboardMode;
  USHORT  NumberOfFunctionKeys;
  USHORT  NumberOfIndicators;
  USHORT  NumberOfKeysTotal;
  ULONG  InputDataQueueLength;
  KEYBOARD_TYPEMATIC_PARAMETERS  KeyRepeatMinimum;
  KEYBOARD_TYPEMATIC_PARAMETERS  KeyRepeatMaximum;
} KEYBOARD_ATTRIBUTES, *PKEYBOARD_ATTRIBUTES;

#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)

#define FILE_ANY_ACCESS                 0
#define METHOD_BUFFERED                 0

#define FILE_DEVICE_KEYBOARD            0x0000000b
#define FILE_DEVICE_MOUSE               0x0000000f

#define IOCTL_MOUSE_QUERY_ATTRIBUTES         CTL_CODE(FILE_DEVICE_MOUSE, 0, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_KEYBOARD_QUERY_ATTRIBUTES      CTL_CODE(FILE_DEVICE_KEYBOARD, 0x0000, METHOD_BUFFERED, FILE_ANY_ACCESS)

#include "../common/win2k_xp_config.h"
#include "../common/win2k_xp_debug.h"
#include "myList.h"

typedef HRESULT (WINAPI *DirectInput8Create_t) (HINSTANCE hinst, DWORD dwVersion, REFIID riidltf, LPVOID * ppvOut, LPUNKNOWN punkOuter);

/*Debugging*/
int debugLevel;
DbgPrintf_t DbgPrintf;

/*Mode select*/
int UseDirectInput;
DirectInput8Create_t dlDirectInput8Create;

/*List of raw input data for each window*/
static myList * rawMouseInputWindowList    = NULL;
static myList * rawKeyboardInputWindowList = NULL;
static myList * rawInputHandleList         = NULL;

/*HOOK HANDLES*/
static HHOOK rawInputMouseHook    = NULL;
static HHOOK rawInputKeyboardHook = NULL;
static BOOL  mouseFirstRun        = TRUE;

/*HANDLE to the heap*/
static HANDLE hHeap = NULL;

//X,Y coords to create relative movement of mouse from low level hook
static LONG mouseHookLastX = 0, mouseHookLastY = 0;

/*DirectInput data*/
LPDIRECTINPUT8         lpdi            = NULL;
LPDIRECTINPUTDEVICE8   m_mouse         = NULL;
LPDIRECTINPUTDEVICE8   m_keyboard      = NULL;
HANDLE                 mouseInputEvent = NULL;

BOOL                   b_registered_mouse;
BOOL                   b_registered_keyboard;
RAWINPUTDEVICE         rid_mouse;
RAWINPUTDEVICE         rid_keyboard;

#define MAX_DEVICE_NAME      32
#define MOUSE_DEVICE_HANDLE  0x1337
#define KBD_DEVICE_HANDLE    0xF001
#define INFO_DEVICE_NAME     "Win2kXPRawInputQuery"

BOOL GetKeyboardDeviceInfo(KEYBOARD_ATTRIBUTES *attrib) {
	HANDLE hDevice;
	DWORD bytesReturned;

	//We really should have this synchronized somehow
	if(!DefineDosDevice(DDD_RAW_TARGET_PATH, INFO_DEVICE_NAME, "\\Device\\KeyboardClass0")) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create Dos Device for kbd 0x%X\n", GetLastError());
		return FALSE;
	}
	
	hDevice = CreateFile("\\\\.\\" INFO_DEVICE_NAME,
		0,
		FILE_SHARE_DELETE|FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		0,
		NULL);
	
	if(hDevice == INVALID_HANDLE_VALUE) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to open Dos Device for kbd 0x%X\n", GetLastError());
		DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
		return FALSE;
	}
	
	if(!DeviceIoControl(
		hDevice,
		IOCTL_KEYBOARD_QUERY_ATTRIBUTES,
		NULL, 0,
		attrib, sizeof(KEYBOARD_ATTRIBUTES),
		&bytesReturned,
		NULL)) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: DevIOCtl failed to query the kbd 0x%X\n", GetLastError());
		CloseHandle(hDevice);
		DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
		return FALSE;
	}
	
	CloseHandle(hDevice);
	DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
	return TRUE;
}

BOOL GetMouseDeviceInfo(MOUSE_ATTRIBUTES *attrib) {
	HANDLE hDevice;
	DWORD bytesReturned;

	//We really should have this synchronized somehow
	if(!DefineDosDevice(DDD_RAW_TARGET_PATH, INFO_DEVICE_NAME, "\\Device\\PointerClass0")) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create Dos Device for mouse 0x%X\n", GetLastError());
		return FALSE;
	}
	
	hDevice = CreateFile("\\\\.\\" INFO_DEVICE_NAME,
		0,
		FILE_SHARE_DELETE|FILE_SHARE_READ|FILE_SHARE_WRITE,
		NULL,
		OPEN_EXISTING,
		0,
		NULL);
	
	if(hDevice == INVALID_HANDLE_VALUE) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to open Dos Device for mouse 0x%X\n", GetLastError());
		DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
		return FALSE;
	}
	
	if(!DeviceIoControl(
		hDevice,
		IOCTL_MOUSE_QUERY_ATTRIBUTES,
		NULL, 0,
		attrib, sizeof(MOUSE_ATTRIBUTES),
		&bytesReturned,
		NULL)) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: DevIOCtl failed to query the mouse 0x%X\n", GetLastError());
		CloseHandle(hDevice);
		DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
		return FALSE;
	}
	
	CloseHandle(hDevice);
	DefineDosDevice(DDD_REMOVE_DEFINITION, INFO_DEVICE_NAME, NULL);
	return TRUE;
}

UINT WINAPI XP_GetRegisteredRawInputDevices(
  PRAWINPUTDEVICE pRawInputDevices, PUINT puiNumDevices, UINT cbSize) {
	UINT nd;
	if(cbSize != sizeof(RAWINPUTDEVICE)) {
		if(DbgPrintf) 
			DbgPrintf(DBG_WARN, 
				S_YELLOW "user32: GetRegisteredRawInputDevices expected structure size %d got %d\n",
				sizeof(RAWINPUTDEVICE), cbSize);
	}
	
	//Since we only can the keyboard and mouse there should only ever be two
	//devices.  This will just return whatever structure they were registered
	//with in this application.  If you register more than once, you get the last.
	nd = (b_registered_mouse?1:0) + (b_registered_keyboard?1:0);
	
	if(nd > *puiNumDevices || !pRawInputDevices) {
		SetLastError(ERROR_INSUFFICIENT_BUFFER);
		*puiNumDevices = nd;
		return -1;
	}
	
	if(b_registered_mouse) {
		memcpy(pRawInputDevices, &rid_mouse, sizeof(RAWINPUTDEVICE));
		pRawInputDevices = (PRAWINPUTDEVICE)((PBYTE)pRawInputDevices + cbSize);
	}
	
	if(b_registered_keyboard) {
		memcpy(pRawInputDevices, &rid_keyboard, sizeof(RAWINPUTDEVICE));
	}
	
	*puiNumDevices = nd;
	return nd;
}

UINT WINAPI XP_GetRawInputBuffer(PRAWINPUT pData, PUINT pcbSize, UINT cbSizeHeader) {
	DWORD count, shuffle = ((sizeof(RAWINPUT)+3) & ~3);
	PRAWINPUT in, out;
	
	GetLock(rawInputHandleList);
	
	if(pData == NULL) {
		//Dunno if this is what is supposed to be returned or not, and if da buffer is to small
		//bad things happen, from the example on msdn that make a 'wild guess' based on this number.
		//WTF!? seriously fucking retarded
		*pcbSize = rawInputHandleList->length * shuffle;
		ReleaseLock(rawInputHandleList);
		return 0;
	}
	
	if((DWORD)pData & 3) {
		//Although the msdn documentation doesn't say anything the NEXTRAWINPUTBLOCK macro implies this is true
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: GetRawInputBuffer misaligned pData parameter\n");
		ReleaseLock(rawInputHandleList);
		return (UINT)-1;
	}
	
	count = 0;
	while(rawInputHandleList->length) {
		in  = (PRAWINPUT)PopFrontAndFree(rawInputHandleList);
		out = (PRAWINPUT)((PBYTE)pData + shuffle);
		memset(out, 0, shuffle);
		
		memcpy(out, in, sizeof(RAWINPUTHEADER));
		memcpy((PBYTE)out + cbSizeHeader, (PBYTE)in + sizeof(RAWINPUTHEADER), sizeof(RAWINPUT) - out->header.dwSize);
		
		HeapFree(hHeap, 0, in);
		count++;
	}
	
	*pcbSize = sizeof(RAWINPUT);
	ReleaseLock(rawInputHandleList);
	return count;
}

UINT WINAPI XP_GetRawInputDeviceInfo(HANDLE hDevice, UINT uiCommand, LPVOID pData, PUINT pcbSize, BOOL wide) {
	UINT size_needed;
	UINT len;
	RID_DEVICE_INFO *info;
	MOUSE_ATTRIBUTES mattrib;
	KEYBOARD_ATTRIBUTES kattrib;
	UINT size_wrote = 0;
	
	if(uiCommand == RIDI_PREPARSEDDATA) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: GetRawInputDeviceInfo - Don't know what to do with RIDI_PREPARSEDDATA\n");
		return 0;
	}
	
	switch(uiCommand) {
		case RIDI_DEVICENAME:
			size_needed = MAX_DEVICE_NAME;
			break;
		case RIDI_DEVICEINFO:
			size_needed = sizeof(RID_DEVICE_INFO);
			break;
		default:
			if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: GetRawInputDeviceInfo - Unknown uiCommand %X\n", uiCommand);
			return 0;
	}
	
	if(!pData) {
		*pcbSize = size_needed;
		return 0;
	}
	
	if(*pcbSize < size_needed) {
		*pcbSize = size_needed;
		return -1;
	}
	
	switch(uiCommand) {
		case RIDI_DEVICENAME:
			switch((DWORD)hDevice) {
				case MOUSE_DEVICE_HANDLE:
					len = strlen("Wrapper Mouse");
					if(!wide) RtlCopyMemory(pData,  "Wrapper Mouse", len*sizeof(CHAR));
					else      RtlCopyMemory(pData, L"Wrapper Mouse", len*sizeof(WCHAR));
					size_wrote = len * (wide ? sizeof(WCHAR) : sizeof(CHAR));
					break;
				case KBD_DEVICE_HANDLE:
					len = strlen("Wrapper Keyboard");
					if(!wide) RtlCopyMemory(pData,  "Wrapper Keyboard", len*sizeof(CHAR));
					else      RtlCopyMemory(pData, L"Wrapper Keyboard", len*sizeof(WCHAR));
					size_wrote = len * (wide ? sizeof(WCHAR) : sizeof(CHAR));
					break;
				default:
					if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: GetRawInputDeviceInfo - Unknown device %X\n", hDevice);
					return 0;
			}
			break;
		case RIDI_DEVICEINFO:
			info = (RID_DEVICE_INFO *)pData;
			info->cbSize = sizeof(RID_DEVICE_INFO);
			size_wrote = sizeof(RID_DEVICE_INFO);
			switch((DWORD)hDevice) {
				case MOUSE_DEVICE_HANDLE:
					info->dwType = RIM_TYPEMOUSE;
					if(GetMouseDeviceInfo(&mattrib)) {
						info->mouse.dwId = mattrib.MouseIdentifier;
						info->mouse.dwNumberOfButtons = mattrib.NumberOfButtons;
						info->mouse.dwSampleRate = mattrib.SampleRate;
						//info->mouse.fHasHorizontalWheel = FALSE;
					}
					break;
				case KBD_DEVICE_HANDLE:
					info->dwType = RIM_TYPEKEYBOARD;
					if(GetKeyboardDeviceInfo(&kattrib)) {
						info->keyboard.dwType    = kattrib.KeyboardIdentifier.Type;
						info->keyboard.dwSubType = kattrib.KeyboardIdentifier.Subtype;
						info->keyboard.dwKeyboardMode = kattrib.KeyboardMode;
						info->keyboard.dwNumberOfFunctionKeys = kattrib.NumberOfFunctionKeys;
						info->keyboard.dwNumberOfIndicators = kattrib.NumberOfIndicators;
						info->keyboard.dwNumberOfKeysTotal = kattrib.NumberOfKeysTotal;
					}
					break;
				default:
					if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: GetRawInputDeviceInfo - Unknown device %X\n", hDevice);
					return 0;
			}
			break;
	}
	
	return size_wrote;
}

UINT WINAPI XP_GetRawInputDeviceInfoA(HANDLE hDevice, UINT uiCommand, LPVOID pData, PUINT pcbSize) {
	return XP_GetRawInputDeviceInfo(hDevice, uiCommand, pData, pcbSize, FALSE);
}

UINT WINAPI XP_GetRawInputDeviceInfoW(HANDLE hDevice, UINT uiCommand, LPVOID pData, PUINT pcbSize) {
	return XP_GetRawInputDeviceInfo(hDevice, uiCommand, pData, pcbSize, TRUE);
}

UINT WINAPI XP_GetRawInputDeviceList(PRAWINPUTDEVICELIST pRawInputDeviceList, PUINT puiNumDevices, UINT cbSize) {
	PRAWINPUTDEVICELIST d;
	
	if(!pRawInputDeviceList) {
		*puiNumDevices = 2;
		return 0;
	}
	
	if(*puiNumDevices < 2) {
		*puiNumDevices = 2;
		SetLastError(ERROR_INSUFFICIENT_BUFFER);
		return (UINT)-1;
	}
	
	d = pRawInputDeviceList;
	d->hDevice = MOUSE_DEVICE_HANDLE;
	d->dwType  = RIM_TYPEMOUSE;
	
	d = (PRAWINPUTDEVICELIST)(((PBYTE)d) + cbSize);
	d->hDevice = KBD_DEVICE_HANDLE;
	d->dwType  = RIM_TYPEKEYBOARD;
	
	*puiNumDevices = 2;
	return 2;
}

BOOL WINAPI XP_SetCursorPos(int X, int Y) {
    mouseHookLastX = X;
    mouseHookLastY = Y;
    return SetCursorPos(X, Y);
}

UINT WINAPI XP_GetRawInputData(HRAWINPUT hRawInput, UINT uiCommand, LPVOID pData, PUINT pcbSize, UINT cbSizeHeader) {
    UINT res = 0;
    RAWINPUT *ri = (RAWINPUT*)hRawInput;
	
	if(cbSizeHeader != sizeof(RAWINPUTHEADER)) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: sizeof(RAWINPUTHEADER) does not match expected\n");
	}
	
	GetLock(rawInputHandleList);

	//Ain't in the list no more
    if(!hRawInput || !IsInListByVal(rawInputHandleList, (DWORD)hRawInput)) {
		ReleaseLock(rawInputHandleList);
        return ((UINT)-1);
	}
    
    if(!pData) {
        switch(uiCommand) {
            case RID_INPUT:
                *pcbSize = sizeof(RAWINPUT);
                res = 0;
                break;
            
            case RID_HEADER:
                *pcbSize = sizeof(RAWINPUTHEADER);
                res = 0;
                break;
            
            default:
                *pcbSize = 0;
                res = (UINT)-1;
        }
    } else {
        switch(uiCommand) {
            case RID_INPUT:
                if(*pcbSize >= sizeof(RAWINPUT)) {
                    res = *pcbSize;
                    RtlCopyMemory(pData, ri, sizeof(RAWINPUT));
                } else {
					res = (UINT)-1;
				}
                break;
            
            case RID_HEADER:
                if(*pcbSize >= sizeof(RAWINPUTHEADER)) {
                    res = *pcbSize;
                    RtlCopyMemory(pData, &ri->header, sizeof(RAWINPUTHEADER));
                } else {
					res = (UINT)-1;
				}
                break;
            
            default:
                res = (UINT)-1;
        }
    }
    
	ReleaseLock(rawInputHandleList);
    return res;
}

VOID AddRawInputToWindowList(RAWINPUT *ri, myList *windowList) {
	myNode * cour = NULL;
	GetLock(rawInputHandleList);
	
	//Remove old rawinputs
	if(rawInputHandleList && rawInputHandleList->length > RAWINPUTHANDLETABLESIZE)
		HeapFree(hHeap, 0, (RAWINPUT *)PopFrontAndFree(rawInputHandleList));

	//Add this new RAWINPUT
    PushBackByVal(rawInputHandleList, (DWORD)ri);

    GetLock(windowList);
    for(cour = windowList->head; cour; cour = cour->next) {
		PostMessageA((HWND)cour->data, WM_INPUT, RIM_INPUT, (LPARAM)ri);
    }				
    ReleaseLock(windowList);
	ReleaseLock(rawInputHandleList);
	return;
}

void SetRawInputMouseHeader(RAWINPUT *ri) {
	ri->header.dwType = RIM_TYPEMOUSE;
    ri->header.dwSize = sizeof(RAWINPUT);
    ri->header.hDevice = MOUSE_DEVICE_HANDLE;
    ri->header.wParam = RIM_INPUT;
	return;	
}

//SetWindowsHookEx WH_MOUSE_LL handler
LRESULT CALLBACK myLowLevelMouseProc(int nCode, WPARAM wParam, LPARAM lParam) {
    static ULONG ulRawButtons = 0;
    RAWMOUSE *rm = NULL;
    RAWINPUT *ri = NULL;
    MSLLHOOKSTRUCT * hs = (MSLLHOOKSTRUCT*)lParam;
	   
    if(nCode == HC_ACTION) { //Should always be true
        ri = (RAWINPUT*)HeapAlloc(hHeap, 0, sizeof(RAWINPUT));
        if(ri) {
			SetRawInputMouseHeader(ri);
			
            rm = &ri->data.mouse;
            if(mouseFirstRun) { //first time around give the absolute position
                rm->usFlags = MOUSE_MOVE_ABSOLUTE;
                rm->lLastX = hs->pt.x;
                rm->lLastY = hs->pt.y;
            } else {
                rm->usFlags = MOUSE_MOVE_RELATIVE;
                if(hs->flags & LLMHF_INJECTED) {
                    rm->lLastX = 0;
                    rm->lLastY = 0;
                } else {
                    rm->lLastX = (LONG)hs->pt.x - mouseHookLastX;
                    rm->lLastY = (LONG)hs->pt.y - mouseHookLastY;
                }
            }
            
            mouseHookLastX = hs->pt.x;
            mouseHookLastY = hs->pt.y;	
            
            rm->ulButtons = 0;
            rm->usButtonData = 0;
            rm->usButtonFlags = 0;
            rm->ulExtraInformation = 0;
        
			//note ulRawButtons is static and we keep track of the state of the buttons this way
            switch(wParam)
            {
                case WM_LBUTTONDOWN:
                    rm->usButtonFlags |= RI_MOUSE_LEFT_BUTTON_DOWN;
                    ulRawButtons |= 0x00000001;
                    break;
        
                case WM_LBUTTONUP:
                    rm->usButtonFlags |= RI_MOUSE_LEFT_BUTTON_UP;
                    ulRawButtons &= ~0x00000001;
                    break;                
                    
                case WM_RBUTTONDOWN:
                    rm->usButtonFlags |= RI_MOUSE_RIGHT_BUTTON_DOWN;
                    ulRawButtons |= 0x00000002;
                    break;             
                    
                case WM_RBUTTONUP:
                    rm->usButtonFlags |= RI_MOUSE_RIGHT_BUTTON_UP;
                    ulRawButtons &= ~0x00000002;
                    break;                

                case WM_MBUTTONDOWN:
                    rm->usButtonFlags |= RI_MOUSE_MIDDLE_BUTTON_DOWN;
                    ulRawButtons |= 0x00000004;
                    break;             
                    
                case WM_MBUTTONUP:
                    rm->usButtonFlags |= RI_MOUSE_MIDDLE_BUTTON_UP;
                    ulRawButtons &= ~0x00000004;
                    break; 
                
                case WM_MOUSEWHEEL:
                    rm->usButtonFlags |= RI_MOUSE_WHEEL;
                    rm->usButtonData = HIWORD(hs->mouseData);
                    break;
					
				case WM_XBUTTONDOWN:
					if(HIWORD(hs->mouseData) == XBUTTON1)      rm->usButtonFlags |= RI_MOUSE_BUTTON_4_DOWN;
					else if(HIWORD(hs->mouseData) == XBUTTON2) rm->usButtonFlags |= RI_MOUSE_BUTTON_5_DOWN;
					ulRawButtons |= (1 << (HIWORD(hs->mouseData) + 2));
					break;
					
				case WM_XBUTTONUP:
					if(HIWORD(hs->mouseData) == XBUTTON1)      rm->usButtonFlags |= RI_MOUSE_BUTTON_4_UP;
					else if(HIWORD(hs->mouseData) == XBUTTON2) rm->usButtonFlags |= RI_MOUSE_BUTTON_5_UP;
					ulRawButtons &= ~(1 << (HIWORD(hs->mouseData) + 2));
					break;
            }
            rm->ulRawButtons = ulRawButtons;
                                    
            if(!mouseFirstRun) {
				AddRawInputToWindowList(ri, rawMouseInputWindowList);
            } else {
				mouseFirstRun = FALSE;
			}
        }
    }
    
    return CallNextHookEx(rawInputMouseHook, nCode, wParam, lParam);
}

LRESULT CALLBACK myLowLevelKeyboardProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    KBDLLHOOKSTRUCT * ks = (KBDLLHOOKSTRUCT*)lParam;    
    RAWKEYBOARD *rk=NULL;
    RAWINPUT *ri=NULL;
    
    if(nCode == HC_ACTION)
    {
        ri = (RAWINPUT*)HeapAlloc(hHeap, 0, sizeof(RAWINPUT));
        if(ri)
        {
            rk = &ri->data.keyboard;
            rk->MakeCode = ks->scanCode;
            rk->Flags = (wParam == WM_KEYUP || wParam == WM_SYSKEYUP) ? RI_KEY_BREAK : RI_KEY_MAKE;
            rk->Reserved = 0;
            rk->VKey = ks->vkCode;
            rk->Message = wParam;
            rk->ExtraInformation = 0;
        
            ri->header.dwType = RIM_TYPEKEYBOARD;
            ri->header.dwSize = sizeof(RAWINPUT);
            ri->header.hDevice = KBD_DEVICE_HANDLE;
            ri->header.wParam = RIM_INPUT;
			
			AddRawInputToWindowList(ri, rawKeyboardInputWindowList);            
        }
    }
    
    return CallNextHookEx(rawInputKeyboardHook, nCode, wParam, lParam);
}

RAWINPUT *GetDirectInputMouseData() {
	static DIMOUSESTATE2 mouse_state;
	static DIMOUSESTATE2 mouse_state_prev;
	RAWINPUT *raw;
	HRESULT hr;
	int i;
	
	raw = (RAWINPUT*)HeapAlloc(hHeap, 0, sizeof(RAWINPUT));
	SetRawInputMouseHeader(raw);
	
	//Try to get input, and maybe reaquire input
	hr = m_mouse->lpVtbl->GetDeviceState(m_mouse, sizeof(DIMOUSESTATE2), (LPVOID)&mouse_state);
	if(FAILED(hr)) {
		m_mouse->lpVtbl->Acquire(m_mouse);
		while(hr == DIERR_INPUTLOST) {
			hr = m_mouse->lpVtbl->Acquire(m_mouse);
		}
		if(FAILED(hr)) {
			HeapFree(hHeap, 0, raw);
			return NULL;
		}
		m_mouse->lpVtbl->GetDeviceState(m_mouse, sizeof(DIMOUSESTATE2), (LPVOID)&mouse_state);
	}
	
	raw->data.mouse.usFlags = MOUSE_MOVE_RELATIVE;
	raw->data.mouse.lLastX = mouse_state.lX;
	raw->data.mouse.lLastY = mouse_state.lY;
	raw->data.mouse.usButtonData = mouse_state.lZ & 0xFFFF;
	raw->data.mouse.usButtonFlags = 0;
	raw->data.mouse.ulRawButtons = 0;
	
	if(raw->data.mouse.usButtonData != 0) raw->data.mouse.usButtonFlags |= RI_MOUSE_WHEEL;
	
	for(i = 0; i < 8; i++) {
		if(mouse_state.rgbButtons[i] & 0x80) {
			raw->data.mouse.ulRawButtons |= 1<<i;
		}
	}
	
	if(mouse_state.rgbButtons[0] & 0x80 && !(mouse_state_prev.rgbButtons[0] & 0x80))
		raw->data.mouse.usButtonFlags |= RI_MOUSE_LEFT_BUTTON_DOWN;
	
	if(!(mouse_state.rgbButtons[0] & 0x80) && mouse_state_prev.rgbButtons[0] & 0x80)
		raw->data.mouse.usButtonFlags |= RI_MOUSE_LEFT_BUTTON_UP;
		
	if(mouse_state.rgbButtons[1] & 0x80 && !(mouse_state_prev.rgbButtons[1] & 0x80))
		raw->data.mouse.usButtonFlags |= RI_MOUSE_RIGHT_BUTTON_DOWN;
	
	if(!(mouse_state.rgbButtons[1] & 0x80) && mouse_state_prev.rgbButtons[1] & 0x80)
		raw->data.mouse.usButtonFlags |= RI_MOUSE_RIGHT_BUTTON_UP;
		
	if(mouse_state.rgbButtons[2] & 0x80 && !(mouse_state_prev.rgbButtons[2] & 0x80))
		raw->data.mouse.usButtonFlags |= RI_MOUSE_MIDDLE_BUTTON_DOWN;
	
	if(!(mouse_state.rgbButtons[2] & 0x80) && mouse_state_prev.rgbButtons[2] & 0x80)
		raw->data.mouse.usButtonFlags |= RI_MOUSE_MIDDLE_BUTTON_UP;
		
	if(mouse_state.rgbButtons[3] & 0x80 && !(mouse_state_prev.rgbButtons[3] & 0x80))
		raw->data.mouse.usButtonFlags |= RI_MOUSE_BUTTON_4_DOWN;
	
	if(!(mouse_state.rgbButtons[3] & 0x80) && mouse_state_prev.rgbButtons[3] & 0x80)
		raw->data.mouse.usButtonFlags |= RI_MOUSE_BUTTON_4_UP;
	
	if(mouse_state.rgbButtons[4] & 0x80 && !(mouse_state_prev.rgbButtons[4] & 0x80))
		raw->data.mouse.usButtonFlags |= RI_MOUSE_BUTTON_5_DOWN;
	
	if(!(mouse_state.rgbButtons[4] & 0x80) && mouse_state_prev.rgbButtons[4] & 0x80)
		raw->data.mouse.usButtonFlags |= RI_MOUSE_BUTTON_5_UP;
		
	memcpy(&mouse_state_prev, &mouse_state, sizeof(DIMOUSESTATE2));
	
	return raw;
}

//Mouse and Keyboard seperate just to simplify things a bit
DWORD WINAPI PollDirectInputMouse(PVOID data) {
	RAWINPUT *raw;
	for(;;) {
		if(WaitForSingleObject(mouseInputEvent, 1000*2) != WAIT_ABANDONED) {
			raw = GetDirectInputMouseData();
			if(raw) AddRawInputToWindowList(raw, rawMouseInputWindowList);
		}
	}
	return 0;
}

VOID DeRegisterDirectInputMouse() {
	m_mouse->lpVtbl->Unacquire(m_mouse);
	m_mouse->lpVtbl->SetEventNotification(m_mouse, NULL);
}

BOOL RegisterDirectInputMouse(PCRAWINPUTDEVICE pRawInputDevices, HWND hWnd) {
	HRESULT hr;
	DWORD flags;
	
	//Try to map these flags to DirectX
	flags = 0;
	if(pRawInputDevices->dwFlags & RIDEV_INPUTSINK) flags |= DISCL_BACKGROUND;
	else                                            flags |= DISCL_FOREGROUND;
	flags |= DISCL_NONEXCLUSIVE;
		
	GetLock(rawMouseInputWindowList);
	//Init DirectInput if nessecary
	if (lpdi == NULL && FAILED(dlDirectInput8Create(GetModuleHandle(NULL), DIRECTINPUT_VERSION, &IID_IDirectInput8, (void**)&lpdi, NULL))) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create DirectInput interface\n");
		ReleaseLock(rawMouseInputWindowList);
		return FALSE;
	}
	
	//Init MouseDevice if nessecary
	if (m_mouse == NULL && FAILED(lpdi->lpVtbl->CreateDevice(lpdi, &GUID_SysMouse, &m_mouse, NULL))) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create DirectInput mouse device\n");
		ReleaseLock(rawMouseInputWindowList);
		return FALSE;
	}
	ReleaseLock(rawMouseInputWindowList);
	
	//Make friendly with this hWnd
	if (FAILED(m_mouse->lpVtbl->SetCooperativeLevel(m_mouse, hWnd, flags))) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to set cooperative level on DirectInput mouse device\n");
		return FALSE;
	}
	
	if (FAILED(m_mouse->lpVtbl->SetDataFormat(m_mouse, &c_dfDIMouse2))) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to set data format on DirectInput mouse device\n");
		return FALSE;
	}

	GetLock(rawMouseInputWindowList);
	if(!mouseInputEvent) {
		mouseInputEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
		if(!mouseInputEvent) {
			if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create event for DirectInput mouse\n");
			ReleaseLock(rawMouseInputWindowList);
			return FALSE;
		}
		
		if(!CreateThread(NULL, 0, PollDirectInputMouse, NULL, 0, NULL)) {
			if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to create polling thread for DirectInput mouse\n");
			ReleaseLock(rawMouseInputWindowList);
			return FALSE;
		}
	}
	
	//Must be NOT be acquired to SetEventNotification
	m_mouse->lpVtbl->Unacquire(m_mouse);
	if(FAILED(m_mouse->lpVtbl->SetEventNotification(m_mouse, mouseInputEvent))) {
		if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to SetEventNotification for DirectInput mouse\n");
		ReleaseLock(rawMouseInputWindowList);
		return FALSE;		
	}
	
	ReleaseLock(rawMouseInputWindowList);
	return TRUE;
}


BOOL WINAPI XP_RegisterRawInputDevices(PCRAWINPUTDEVICE pRawInputDevices, UINT uiNumDevices, UINT cbSize) {
    int i;
	HWND hWnd;

    if(!pRawInputDevices || !uiNumDevices || !cbSize) {
        SetLastError(ERROR_INVALID_PARAMETER);
        return FALSE;
    }
	
    for(i=0; i<uiNumDevices; i++) {
		//Get the window handle if we need to
		hWnd = pRawInputDevices[i].hwndTarget;
		if(!hWnd) hWnd = GetFocus();
		if(!hWnd) hWnd = GetForegroundWindow(); //try this too
		if(!hWnd) {
			if(DbgPrintf) DbgPrintf(DBG_WARN,  "user32: couldn't get HWND for RegisterRawInputDevices\n");
			return FALSE;
		}
	
        // Mouse input
        if(pRawInputDevices[i].usUsagePage==1 && pRawInputDevices[i].usUsage==2) { //Mouse should match this
			GetLock(rawMouseInputWindowList);
            if(!(pRawInputDevices[i].dwFlags & RIDEV_REMOVE)) {
				//Add			
				if(DbgPrintf) DbgPrintf(DBG_TRACE, "user32: Adding a RAWINPUT mouse\n");
				
                if(pRawInputDevices[i].dwFlags & RIDEV_CAPTUREMOUSE)
                    SetCapture(hWnd);
				
				//Add the hWnd to the list
                if(!IsInListByVal(rawMouseInputWindowList, (DWORD)hWnd))
                    PushFrontByVal(rawMouseInputWindowList, (DWORD)hWnd);
                
				//If not getting low level data yet activate
				if(UseDirectInput) {
					if(!RegisterDirectInputMouse(&pRawInputDevices[i], hWnd) && DbgPrintf)
						DbgPrintf(DBG_ERROR, S_RED "user32: Could not register DirectInput mouse\n");				
                } else if(!rawInputMouseHook) {
                    rawInputMouseHook = SetWindowsHookExA(WH_MOUSE_LL, myLowLevelMouseProc, GetModuleHandle("user32.dll"), 0);
					if(!rawInputMouseHook && DbgPrintf)
						DbgPrintf(DBG_ERROR, S_RED "user32: Could not SetWindowsHookEx WH_MOUSE_LL LastError=0x%X\n", GetLastError());
				}
				
				//Save this structure to return with GetRegisteredRawInputDevices
				b_registered_mouse = TRUE;
				memcpy(&rid_mouse, &pRawInputDevices[i], sizeof(RAWINPUTDEVICE));
            } else {
				//Remove
				if(DbgPrintf) DbgPrintf(DBG_TRACE, "user32: Removing a RAWINPUT mouse\n");
				
                if(pRawInputDevices[i].hwndTarget) {
					//This actually is not allowed
                    RemoveNodeByVal(rawMouseInputWindowList, (DWORD)pRawInputDevices[i].hwndTarget);
                } else {
					//Remove dem all!
                    while(rawMouseInputWindowList->length)
                        PopFrontAndFree(rawMouseInputWindowList);
                }
                
                ReleaseCapture();
                
				if(rawMouseInputWindowList->length==0) {
					if(UseDirectInput) {
						DeRegisterDirectInputMouse();
					} else if(rawInputMouseHook) {
						UnhookWindowsHookEx(rawInputMouseHook);
						rawInputMouseHook = NULL;
					}
				}
				
				b_registered_mouse = FALSE;
            }
			ReleaseLock(rawMouseInputWindowList);
        }
          
        // Keyboard input
        if(pRawInputDevices[i].usUsagePage==1 && pRawInputDevices[i].usUsage==6) { //kbd should match this
			GetLock(rawKeyboardInputWindowList);
            if(!(pRawInputDevices[i].dwFlags & RIDEV_REMOVE)) {
				//Add			
				if(DbgPrintf) DbgPrintf(DBG_TRACE, "user32: Adding a RAWINPUT keyboard\n");
				
                if(!IsInListByVal(rawKeyboardInputWindowList, (DWORD)hWnd))
                    PushFrontByVal(rawKeyboardInputWindowList, (DWORD)hWnd);
					
				if(UseDirectInput)
					if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: no DirectInput support for KBD yet, defaulting to Low Level Hook\n");
                
                if(!rawInputKeyboardHook)
                    rawInputKeyboardHook = SetWindowsHookExA(WH_KEYBOARD_LL, myLowLevelKeyboardProc, GetModuleHandle("user32.dll"), 0);
					
				//Save this structure to return with GetRegisteredRawInputDevices
				b_registered_keyboard = TRUE;
				memcpy(&rid_keyboard, &pRawInputDevices[i], sizeof(RAWINPUTDEVICE));
            } else {
				//Remove
				if(DbgPrintf) DbgPrintf(DBG_TRACE, "user32: Removing a RAWINPUT keyboard\n");
				
                if(pRawInputDevices[i].hwndTarget) {
					//This actually is not allowed
                    RemoveNodeByVal(rawKeyboardInputWindowList, (DWORD)pRawInputDevices[i].hwndTarget);
                } else {
					//Remove dem all!
                    while(rawKeyboardInputWindowList->length)
                        PopFrontAndFree(rawKeyboardInputWindowList);
                }
                
                if(rawInputKeyboardHook && rawKeyboardInputWindowList->length==0) {
                    UnhookWindowsHookEx(rawInputKeyboardHook);
                    rawInputKeyboardHook = NULL;
                }
				
				b_registered_keyboard = FALSE;
            }
			ReleaseLock(rawKeyboardInputWindowList);
        }
    }
    
    return TRUE;
}

LRESULT XP_DefRawInputProc(PRAWINPUT *paRawInput, INT nInput, UINT cbSizeHeader) {
	/*
	The DefRawInputProc function calls the default raw input procedure to
	provide default processing for any raw input messages that an application
	does not process. This function ensures that every message is processed.
	DefRawInputProc is called with the same parameters received by the window
	procedure. 
	*/
	return S_OK;
}


BOOL WINAPI XP_GetLayeredWindowAttributes(HWND hwnd, COLORREF *pcrKey, BYTE *pbAlpha, DWORD *pdwFlags) {
	if(!(GetWindowLong(hwnd, GWL_EXSTYLE) & WS_EX_LAYERED)) return FALSE;
	if(pcrKey) {
		*pcrKey = 0x00FFFFFF;
	}
	if(pbAlpha) {
		*pbAlpha = 0xFF;
	}
	if(pdwFlags) {
		*pdwFlags = LWA_ALPHA;
	}
	return TRUE;
}

BOOL WINAPI XP_PrintWindow(HWND hwnd, HDC  hdcBlt, UINT nFlags) {
	HDC dc;
	RECT rect;
	LONG width, height;
	BOOL ret;
	
	if(nFlags & PW_CLIENTONLY) dc = GetDC(hwnd);
	else                       dc = GetWindowDC(hwnd);
	if(dc == NULL) return FALSE;
	
	if(nFlags & PW_CLIENTONLY) ret = GetClientRect(hwnd, &rect);
	else                       ret = GetWindowRect(hwnd, &rect);
	
	if(ret) {
		width  = RECTWIDTH(rect);
        height = RECTHEIGHT(rect);
		
		ret = BitBlt(hdcBlt, 0, 0, width, height, dc, 0, 0, SRCCOPY);
		ReleaseDC(hwnd, dc);
		return ret;
	} else {
		ReleaseDC(hwnd, dc);
		return FALSE;
	}
}

int WINAPI XP_GetWindowRgnBox(HWND hWnd, LPRECT lprc)
{
    HRGN VisRgn = NULL, trgn = NULL;
    RECT wRect;
    int Ret = ERROR, tret = 0;
    RGNDATA * rgn = NULL;
        
    if(!hWnd)
        return ERROR;
    
    if(!lprc)
       return ERROR;
        
    if(GetWindowRect(hWnd, &wRect))
    {
        VisRgn = CreateRectRgnIndirect(&wRect);
        OffsetRgn(VisRgn, -wRect.left, -wRect.top);
        
        if(GetWindowRgn(hWnd, trgn) != ERROR && !(GetWindowLongA(hWnd, GWL_STYLE) & WS_MINIMIZE))
        {
            CombineRgn(VisRgn, VisRgn, trgn, RGN_AND);
            DeleteObject(trgn);
        }
        
        tret = GetRegionData(VisRgn, 0, NULL);
        rgn = HeapAlloc(hHeap, 0, tret);
        GetRegionData(VisRgn, tret, rgn);
        
        Ret = rgn->rdh.iType;
        *lprc = rgn->rdh.rcBound;
        
        HeapFree(hHeap, 0, rgn);
    }
    
    DeleteObject(VisRgn);
    
    return Ret;
}

void WINAPI XP_DisableProcessWindowsGhosting(VOID) {
	/*
	  Disables the window ghosting feature for the calling graphical user interface (GUI) process.
	  Window ghosting is a Windows Manager feature that lets the user minimize, move, or close the
	  main window of an application that is not responding.
	 */
	return;
}

BOOL WINAPI XP_SetProcessDPIAware(VOID) {
	return TRUE;
}

HWND
WINAPI
XP_CreateWindowExA(
  DWORD dwExStyle,
  LPCSTR lpClassName,
  LPCSTR lpWindowName,
  DWORD dwStyle,
  int X,
  int Y,
  int nWidth,
  int nHeight,
  HWND hWndParent,
  HMENU hMenu,
  HINSTANCE hInstance,
  LPVOID lpParam) {
	if(dwExStyle & WS_EX_COMPOSITED) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: CreateWindowEx ignoring WS_EX_COMPOSITED\n");
	}  
	dwExStyle &= ~WS_EX_COMPOSITED;
	return CreateWindowExA(dwExStyle, lpClassName, lpWindowName, dwStyle, X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam);
}

HWND
WINAPI
XP_CreateWindowExW(
  DWORD dwExStyle,
  LPCWSTR lpClassName,
  LPCWSTR lpWindowName,
  DWORD dwStyle,
  int X,
  int Y,
  int nWidth,
  int nHeight,
  HWND hWndParent,
  HMENU hMenu,
  HINSTANCE hInstance,
  LPVOID lpParam) {
	if(dwExStyle & WS_EX_COMPOSITED) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: CreateWindowEx ignoring WS_EX_COMPOSITED\n");
	}  
	dwExStyle &= ~WS_EX_COMPOSITED;
	return CreateWindowExW(dwExStyle, lpClassName, lpWindowName, dwStyle, X, Y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam);
}

#define CS_DROPSHADOW       0x00020000
ATOM WINAPI XP_RegisterClassA(CONST WNDCLASSA *lpWndClass) {
	if(lpWndClass->style & CS_DROPSHADOW) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: RegisterClass ignoring CS_DROPSHADOW\n");
	}
	((WNDCLASSA *)lpWndClass)->style &= ~CS_DROPSHADOW;
	return RegisterClassA(lpWndClass);
}

ATOM WINAPI XP_RegisterClassW(CONST WNDCLASSW *lpWndClass) {
	if(lpWndClass->style & CS_DROPSHADOW) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: RegisterClass ignoring CS_DROPSHADOW\n");
	}
	((WNDCLASSW *)lpWndClass)->style &= ~CS_DROPSHADOW;
	return RegisterClassW(lpWndClass);
}

ATOM WINAPI XP_RegisterClassExA(CONST WNDCLASSEXA *lpwcx) {
	if(lpwcx->style & CS_DROPSHADOW) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: RegisterClassEx ignoring CS_DROPSHADOW\n");
	}
	((WNDCLASSEXA *)lpwcx)->style &= ~CS_DROPSHADOW;
	return RegisterClassExA(lpwcx);
}

ATOM WINAPI XP_RegisterClassExW(CONST WNDCLASSEXW *lpwcx) {
	if(lpwcx->style & CS_DROPSHADOW) {
		if(DbgPrintf) DbgPrintf(DBG_WARN, S_YELLOW "user32: RegisterClassEx ignoring CS_DROPSHADOW\n");
	}
	((WNDCLASSEXW *)lpwcx)->style &= ~CS_DROPSHADOW;
	return RegisterClassExW(lpwcx);
}

BOOL WINAPI XP_IsWinEventHookInstalled(DWORD event) {
	//Why would you care?
	if(DbgPrintf) DbgPrintf(DBG_INFO, "user32: IsWinEventHookInstalled stub\n");
	return TRUE;
}

BOOL WINAPI DllMain(
  HINSTANCE hinstDLL,
  DWORD fdwReason,
  LPVOID lpvReserved) {
	static char buffer[MAX_PATH];
	static char ini_filename[] = INI_FILENAME;
	DWORD read;
	DebugLevel_t DebugLevel;
	HMODULE hMod;
	
  	switch(fdwReason) {
		case DLL_PROCESS_ATTACH:
			if(!hHeap) hHeap = GetProcessHeap();
			
			//Init debug data
			DbgPrintf = GetDbgPrintf();
			DebugLevel = GetDebugLevel();
			
			if(DebugLevel) debugLevel = DebugLevel();
			else           debugLevel = 0;
			
			if(DbgPrintf) DbgPrintf(DBG_ALWAYS, "user32: USER32.DLL Wrapper Init\n");
			
			//Get INI file path
			read = GetCurrentDirectoryA(sizeof(buffer), buffer);
			if(read == 0 || read + sizeof(ini_filename) >= sizeof(buffer)) break;
			lstrcatA(buffer, ini_filename);
			
			UseDirectInput = GetPrivateProfileIntA(USER32_SECTION, "UseDirectInput", 0, buffer);
			if(DbgPrintf && UseDirectInput) DbgPrintf(DBG_INFO, "user32: Using direct input for raw input\n");
			
			if(UseDirectInput) {
				hMod = LoadLibraryA("DINPUT8.DLL");
				if(!hMod) {
					if(DbgPrintf) DbgPrintf(DBG_ERROR, S_RED "user32: Failed to load DINPUT8.DLL, reverting it Low Level hook\n");
					UseDirectInput = 0;
				} else {
					//Should never fail
					dlDirectInput8Create = (DirectInput8Create_t)GetProcAddress(hMod, "DirectInput8Create");
				}
			}
			
			if(!rawMouseInputWindowList) InitializeList(&rawMouseInputWindowList);
			if(!rawKeyboardInputWindowList) InitializeList(&rawKeyboardInputWindowList);
			if(!rawInputHandleList) InitializeList(&rawInputHandleList);
			
			break;
	}
	
	return TRUE;
}

